#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include "sareadline.h"
#include "strlist.h"
#include "tokenizer.h"
#include <strings.h> // strncmp, strlen

extern char **environ;

int exit_builtin(strlist* tokens) {
	exit(0);
	return 0; // wird nicht erreicht, ist aber syntaktisch korrekt
}

int echo_builtin(strlist* tokens) {
	char first = 1;
	for(int i = 1; i<tokens->len; i++) {
		printf("%s%s", first?"":" ", tokens->list[i]);
		first = 0;
	}
	printf("\n");
	return 0; 
}

int env_builtin(strlist* tokens) {
	char **env2 = environ;
	while(*env2) 
		printf("%s\n", *env2++);
	return 0; 
}

int setenv_builtin(strlist* tokens) {
	if(tokens->len < 2) {
		printf("setenv: need a new environment variable!\n");
		return 1;
	}
	int len = strlen(tokens->list[1]);
	char *newstr = malloc(sizeof(char)*len);
	strncpy(newstr, tokens->list[1], len);
	// Damit putenv gefunden wird, muss das Makro __EXTENSIONS__
	// definiert sein. Auf der gcc-Kommandozeile geht dies mit
	// -D__EXTENSIONS__ - in den Makefiles hier wurde das bereits
	// eingebaut.
	return putenv(newstr); 
}

struct builtin {
	char *cmdname;
	int (*function)(strlist* tokens);
};

struct builtin exitbuiltin = {
	"exit",
	&exit_builtin
};

struct builtin echobuiltin = {
	"echo",
	&echo_builtin
};

struct builtin envbuiltin = {
	"env",
	&env_builtin
};

struct builtin setenvbuiltin = {
	"setenv",
	&setenv_builtin
};

struct builtin *builtins[] = {
	&exitbuiltin,
	&echobuiltin,
	&envbuiltin,
	&setenvbuiltin,
	NULL
};



/*
	assign an opened file with the given flags and mode to fd
*/
void fassign ( int fd , char* path , int oflags , mode_t mode ) {
	int newfd = open( path , oflags , mode);
	if ( newfd < 0) {
		perror ( path ); exit (255);
	}
	if ( dup2(newfd , fd ) < 0) {
		perror ( "dup2" ); exit (255);
	}
	close (newfd );
}

int main () {
	stralloc line = {0};
	while ( printf ( "%% "), readline ( stdin , &line )) {
		strlist tokens = {0};
		stralloc_0 (&line ); /* required by tokenizer () */
		if (! tokenizer (&line , & tokens )) break;
		if ( tokens.len == 0) continue;

		
		struct builtin *builtin_to_execute = NULL;
		struct builtin *b = NULL;
		// Suche nach passenden Builtins
		for(int i = 0; (b = builtins[i]) != NULL; i++) 
			if(strncmp(tokens.list[0], b->cmdname, strlen(b->cmdname)) == 0) {
				builtin_to_execute = b;
				break;
			}
		// Falls ein Builtin gefunden wurde, fuehren wir es aus	
		if(builtin_to_execute != NULL) {
			printf("executing builtin %s\n", b->cmdname);
			int returnvalue = b->function(&tokens);
			printf("builtin %s returned value %d\n", b->cmdname, returnvalue);
			continue; // nicht mehr forken, wir sind fertig
		}

		pid_t child = fork ();
		if ( child == -1) {
			perror ( " fork" ); continue;
		}
		if ( child == 0) {
			strlist argv = {0}; /* list of arguments */
			char* cmdname = 0; /* first argument */
			char* path ; /* of output files */
			int oflags ;
	
			for ( int i = 0; i < tokens.len ; ++ i ) {
				switch ( tokens.list [ i ][0]) {
					case '<':
						fassign (0, & tokens.list [ i ][1], O_RDONLY, 0);
						break;
					case '>' :
						path = &tokens.list [ i ][1];
						oflags = O_WRONLY|O_CREAT;
						if (*path == '>' ) {
							++path ; oflags |= O_APPEND;
						} else {
							oflags |= O_TRUNC;
						}
						fassign (1, path , oflags , 0666);
						break;
					default :
						strlist_push (&argv , tokens .list[ i ]);
						if ( cmdname == 0) cmdname = tokens.list[ i ];
				}
			}
			if ( cmdname == 0) exit (0);
			strlist_push(&argv, 0);
			execvp (cmdname, argv.list );
			perror (cmdname);
			exit (255);
		}

	/* wait for termination of child */
	int stat ;
	pid_t pid = wait(&stat );
	if ( pid == child ) {
		if (WIFEXITED(stat)) {
			int code = WEXITSTATUS(stat);
			if ( code && code != 255) {
				printf ( "terminated with exit code %d\n", code);
			}
		} else {
			printf ( "terminated abnormally\n");
		}
	} else {
		perror ( "wait" );
	}
	}
}
